#include "../tools/tools.h"

#include <EGrabber.h>
#include <RGBConverter.h>
#include <string.h> // memcpy

using namespace Euresys;

namespace {

class Sublink;

class Callbacks {
    public:
        virtual void onNewBufferEvent(Sublink &sublink, const NewBufferData &data)      {};
        virtual void onIoToolboxEvent(Sublink &sublink,const IoToolboxData &data)       {};
        virtual void onCicEvent(Sublink &sublink,const CicData &data)                   {};
        virtual void onDataStreamEvent(Sublink &sublink,const DataStreamData &data)     {};
        virtual void onCxpInterfaceEvent(Sublink &sublink,const CxpInterfaceData &data) {};
};

typedef EGrabber<CallbackOnDemand> EGrabberT;

class Sublink: public EGrabberT {
    public:
        Sublink(Callbacks &callbacks, EGenTL &gentl, int interfaceIndex = 0, int deviceIndex = 0, int dataStreamIndex = 0)
        : EGrabberT(gentl, interfaceIndex, deviceIndex, dataStreamIndex)
        , callbacks(callbacks)
        {}
    private:
        Callbacks &callbacks;
        void onNewBufferEvent(const NewBufferData &data) {
            callbacks.onNewBufferEvent(*this, data);
        }
        void onIoToolboxEvent(const IoToolboxData &data) {
            callbacks.onIoToolboxEvent(*this, data);
        }
        void onCicEvent(const CicData &data) {
            callbacks.onCicEvent(*this, data);
        }
        void onDataStreamEvent(const DataStreamData &data) {
            callbacks.onDataStreamEvent(*this, data);
        }
        void onCxpInterfaceEvent(const CxpInterfaceData &data) {
            callbacks.onCxpInterfaceEvent(*this, data);
        }
};

#ifdef _MSC_VER
#pragma warning( push )
#pragma warning( disable : 4355 ) // 'this' : used in base member initializer list
#endif
class SublinkManager: private Callbacks {
    public:
        SublinkManager(EGenTL &gentl, size_t BUFFER_COUNT, int interfaceMaster, int interfaceSlave)
        : BUFFER_COUNT(BUFFER_COUNT)
        , currentFrame(0)
        , rgbConverter(gentl)
        , master(*this, gentl, interfaceMaster)
        , slave(*this, gentl, interfaceSlave)
        {
            master.runScript(Tools::getScriptPath("setup"));
            slave.runScript(Tools::getScriptPath("setup"));
            // Read Sublink Master settings
            width = master.getWidth();
            height = master.getHeight();
            pixelFormat = master.getPixelFormat();
            payloadSize = master.getPayloadSize();
            // Setup Sublink Slave
            slave.setInteger<RemoteModule>("Width", static_cast<int64_t>(width));
            slave.setInteger<RemoteModule>("Height", static_cast<int64_t>(height));
            slave.setString<RemoteModule>("PixelFormat", pixelFormat);
            // Prepare buffers
            pitch = payloadSize / height;
            for (size_t i = 0; i < BUFFER_COUNT; ++i) {
                buffers.push_back(static_cast<uint8_t *>(malloc(payloadSize)));
                master.announceAndQueue(GenTLMemory(payloadSize / 2));
                slave.announceAndQueue(GenTLMemory(payloadSize / 2));
            }
        }
        virtual ~SublinkManager() {
            for (std::vector<uint8_t *>::iterator it = buffers.begin(); it != buffers.end(); ++it) {
                free(*it);
            }
        }
        void go() {
            slave.start(BUFFER_COUNT);
            master.start(BUFFER_COUNT);
            const uint64_t timeout = 1000;
            for (size_t i = 0; i < BUFFER_COUNT; ++i) {
                currentFrame = i;
                master.processEvent<NewBufferData>(timeout);
                slave.processEvent<NewBufferData>(timeout);
            }
            for (size_t i = 0; i < BUFFER_COUNT; ++i) {
                RGBConverter::BGR8 bgr(rgbConverter, buffers[i], pixelFormat, width, height);
                bgr.saveToDisk(Tools::getEnv("sample-output-path") + "/merged.NNN.jpeg", i);
            }
        }
    private:
        bool isMaster(const Sublink &sublink) const {
            return &sublink == &master;
        }
        void mergeIntoBuffer(uint8_t *fromBuffer, bool master, uint8_t *toBuffer) {
            const size_t STRIPE_HEIGHT = 4;
            const size_t STRIPE_SIZE = pitch * STRIPE_HEIGHT;
            const size_t SIZE = payloadSize / 2;
            if (!master) {
                toBuffer += STRIPE_SIZE;
            }
            for (size_t offset = 0; offset < SIZE; offset += STRIPE_SIZE) {
                memcpy(&toBuffer[2 * offset], &fromBuffer[offset], STRIPE_SIZE);
            }
        }        
        void onNewBufferEvent(Sublink &sublink, const NewBufferData &data) {
            ScopedBuffer buffer(sublink, data);
            if (isMaster(sublink)) {
                uint64_t frameId = buffer.getInfo<uint64_t>(gc::BUFFER_INFO_FRAMEID);
                Tools::log("Master got frame " + Tools::toString(frameId));
            }
            uint8_t *imagePointer = buffer.getInfo<uint8_t *>(gc::BUFFER_INFO_BASE);
            mergeIntoBuffer(imagePointer, isMaster(sublink), buffers[currentFrame]);
        };

        const size_t BUFFER_COUNT;
        size_t currentFrame;
        RGBConverter rgbConverter;
        Sublink master;
        Sublink slave;
        std::vector<uint8_t *> buffers;
        
        size_t width;
        size_t height;
        std::string pixelFormat;
        size_t payloadSize;
        size_t pitch;
};
#ifdef _MSC_VER
#pragma warning( pop )
#endif

class AutoTLHandle {
    public:
        AutoTLHandle(EGenTL &gentl)
        : gentl(gentl), h(gentl.tlOpen()) {}
        ~AutoTLHandle() { gentl.tlClose(h); }
        operator gc::TL_HANDLE() { return h; }
    private:
        EGenTL &gentl;
        gc::TL_HANDLE h;
};

int findInterface(EGenTL &gentl, const std::string &name) {
    Tools::log("Looking for grabber '" + name + "'");
    AutoTLHandle tlh(gentl);
    gentl.tlUpdateInterfaceList(tlh);
    uint32_t n = gentl.tlGetNumInterfaces(tlh);
    for (uint32_t i = 0; i < n; ++i) {
        std::string id(gentl.tlGetInterfaceID(tlh, i));
        std::string info(" - [" + Tools::toString(i) + "]='" + id + "'");
        if (id.find(name) != std::string::npos) {
            Tools::log(info + " OK");
            return static_cast<int>(i);
        }
        Tools::log(info);
    }
    throw std::runtime_error("Grabber '" + name + "' not found");
}

}

static void sample() {
    EGenTL genTL;
    int interfaceMaster = findInterface(genTL, "Coaxlink Quad G3 (1-slm-camera)");
    int interfaceSlave  = findInterface(genTL, "Coaxlink Quad G3 (1-sls-camera)");
    const size_t BUFFER_COUNT = 10;
    SublinkManager sublinkManager(genTL, BUFFER_COUNT, interfaceMaster, interfaceSlave);
    sublinkManager.go();
}

static Tools::Sample addSample(__FILE__, "Merge buffers from 2 Sub-Link grabbers on the same PC", sample);
